package com.anabas.ibus;

import java.util.Vector;

import com.anabas.gxo.*;
import com.anabas.util.misc.LogManager;

import javax.jms.*;
import ch.softwired.jms.*;

/**
 * Title:
 * Description:
 * Copyright:    Copyright (c) 2000
 * Company:
 * @author
 * @version 1.0
 */

public class GMS_SubscriberImpl implements GMS_StreamSubscriber {

  private GMS_MessageSelector m_defaultSelector;
  protected TopicSubscriber   m_subscriber;
  protected TopicSession      m_session;
  protected GMS_MessageSelector m_selector;  // The selector for this subscriber.  could be null.
  private GMS_MessageListener m_listener;

  public GMS_SubscriberImpl(TopicSubscriber sub,TopicSession session,GMS_MessageSelector selector) {
    m_subscriber = sub;
    m_selector = selector;
    m_session  = session;
  }

  // Convert a GMS_Message to a JMS Message
  public static GMS_Message convertMessage(Message m) throws GXO_Exception {
    GMS_Message msg;

    if (m instanceof TextMessage) {
      msg = new GMS_TextMessageImpl((TextMessage)m);
    } else if (m instanceof BytesMessage) {
      msg = new GMS_BytesMessageImpl((BytesMessage)m);
    } else if (m instanceof ObjectMessage) {
      msg = new GMS_ObjectMessageImpl((ObjectMessage)m);
    } else {
      throw new GXO_MsgBodyUnsupportedException(m);
    }
    return msg;
  }

  /**
   * Sets the message filter for the subscriber.
   * Best if the filter is implemented on the server side.
   */
  public void setSelector(GMS_MessageSelector s) throws GXO_Exception {
    try {
      if (m_defaultSelector != null) {
        m_selector = m_defaultSelector.mergeWithSelector(s);
      } else {
        m_selector = s;
      }
    } catch (GXO_ConflictingSelectorException e) {
      throw new GXO_AccessDeniedException();
    }
  }

  /**
   * Set a listener for asychronous reception of messages.
   */
  public void setMessageListener(GMS_MessageListener m) throws GXO_Exception {
    try {
      m_subscriber.setMessageListener(new Receiver(this,m));
    } catch (JMSException e) {
      LogManager.err("Subscriber","Unable to set message listener",e);
    }
  }

  /**
   * Synchronous reception of messages.
   *
   * @return null if an error occurs.
   */
  public GMS_Message receiveMessage() throws GXO_MessageException {
    throw new NoSuchMethodError();
  }
}

/**
 * A message listener that handles request messages.  It will send back the
 * reply message with a JMSCorrelationID() equal to the message id of the
 * request to the JMSReplyTo() header field.
 *
 * ToDo:  For some reason, the set delivery mode to NON_PERSISTENT doesn't
 * work here.
 */
class Receiver implements GMS_MessageReplyer, MessageListener {

  private GMS_MessageListener m_listener;
  private GMS_SubscriberImpl  m_parent;
  private TopicPublisher      m_publisher; // For replies.
  private MessageListener     m_listenerAdapter; // Adapts m_listener to javax.jms.MessageListener

  Receiver(GMS_SubscriberImpl parent,GMS_MessageListener l) throws JMSException {
    m_parent = parent;
    m_listener = l;
    TopicSession session = IBusJMSContext.createTopicSession(false,TopicSession.AUTO_ACKNOWLEDGE);
    m_publisher = session.createPublisher(m_parent.m_subscriber.getTopic());
    m_listenerAdapter = new IBusMessageListenerAdapter(l,parent);

    m_publisher.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
  }

  public void reply(GMS_Message original,GMS_Message reply) throws GXO_Exception {
    try {
      Message originalM = GMS_PublisherImpl.convertMessage(original);
      Message replyM    = GMS_PublisherImpl.convertMessage(reply);
      Destination replyTo = originalM.getJMSReplyTo();

      replyM.setJMSCorrelationID(originalM.getJMSMessageID());
      // Here it destination doesn't matter since we are sending it to
      // a temporary topic anyways.
      replyM.setStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY,
                               GXO_JMSConstants.GXO_DESTINATION_ALL);
      m_publisher.publish((Topic)replyTo,replyM);

      LogManager.log(LogManager.ALL,"Receiver","Sent reply [dst=$,correl="+
                     originalM.getJMSMessageID()+"]: " + replyM);
    } catch (JMSException e) {
      throw new GXO_JMSException(e);
    }
  }

  public void onMessage(Message m) {
   try {
      if (m.getJMSReplyTo() == null) {
        // if the message body type is Text, then only log it
        if(m instanceof GMS_TextMessage) {
          LogManager.log(LogManager.ALL,"Receiver","Got Message: [" +
              m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+m);
        }
        m_listenerAdapter.onMessage(m);
      } else {
        // I should block requests from being delivered to myself to prevent
        // self answered requests.
        // if the message body type is Text, then only log it
        if(m instanceof GMS_TextMessage) {
          LogManager.log(LogManager.ALL,"Receiver","Got Request: [" +
                m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+
                            "] [correlation: "+m.getJMSCorrelationID()+"]"+m);
        }
        m_listener.onRequest(GMS_SubscriberImpl.convertMessage(m),this);
      }
    } catch (Exception e) {
      LogManager.err("Subscriber","Received unsupported message type",e);
    }
  }
}


/** ------------------------------------------------------------------------ */
/** ------------------------------------------------------------------------ */
/** -----------------------Not Used Code (was testing)---------------------- */
/** ------------------------------------------------------------------------ */
/** ------------------------------------------------------------------------ */

/**
 * HACK ALERT
 *
 * No choice....can't really use the IBusTopicReplyer class with this so creating
 * a new subscriber that only listens to request messages and use the built in
 * reply(TopicPublisher,original,reply) method to generate reply.
 *
 * Not too efficient as there is 1 ibus subscriber for normal messages and 1 ibus suscriber
 * for request messages for every 1 GMS_StreamSubscriber.
 */
class AsyncReplyerThread extends IBusTopicReplyer implements GMS_MessageReplyer, Runnable {

  private GMS_MessageListener m_listener;
  private GMS_SubscriberImpl  m_parent;
  private TopicSubscriber     m_subscriber; // For receiving requests.
  private TopicPublisher      m_publisher; // For replies.
  private Topic               m_topic;

  AsyncReplyerThread(GMS_SubscriberImpl parent,GMS_MessageListener l) throws JMSException {
    super(parent.m_subscriber.getTopic(),
          IBusJMSContext.createTopicSession(false,TopicSession.AUTO_ACKNOWLEDGE));
    m_parent = parent;
    m_listener = l;
    m_topic = parent.m_subscriber.getTopic();
    if (parent.m_selector != null) {
      m_subscriber = getSession().createSubscriber(m_topic,
                                            parent.m_selector.toString(),false);
    } else {
      m_subscriber = getSession().createSubscriber(m_topic);
    }
    m_publisher = getSession().createPublisher(m_topic);
  }

  public void reply(GMS_Message original,GMS_Message reply) throws GXO_Exception {
    try {
      Message originalM = GMS_PublisherImpl.convertMessage(original);
      Message replyM    = GMS_PublisherImpl.convertMessage(reply);
      // Set this property so clients can select on this and choose to ignore
      // its own messages.  This is really a work around to IBus's noLocal not
      // working.....I think.
      replyM.setStringProperty(GXO_JMSConstants.IBUS_CLIENTID_PROPERTY,IBusJMSContext.getTopicConnection().getClientID());

      reply(m_publisher,originalM,replyM);

      LogManager.log(LogManager.ALL,"ReplyerThread","Sent reply " + replyM);
    } catch (JMSException e) {
      throw new GXO_JMSException(e);
    }
  }

  /**
   * For some reason, onRequest also receives non-request messages.
   */
  public void onRequest(Message m) {
    try {
      if (m.getJMSReplyTo() != null) {
        LogManager.log(LogManager.ALL,"Replyer-onRequest","Got Request: [" +
              m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+"] "+m);
        m_listener.onRequest(GMS_SubscriberImpl.convertMessage(m),this);
      }
    } catch (Exception e) {
      LogManager.err("Subscriber","Received unsupported message type",e);
    }
  }

  public void run() {
   while(true) {
      try {
        Message m = m_subscriber.receive();

        if (m.getJMSReplyTo() != null) {  // Handle only request messages
          //LogManager.log(LogManager.ALL,"Replyer","Got Request: [" +
          // m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+"] "+m);
          //onRequest(m);
        }
      } catch (Exception e) {
        LogManager.err("Subscriber Listner","Error receiving",e);
      }
    }
  }
}

/**
 * HACK ALERT
 *
 * Manually handles request/reply messages as well as regular messages
 *
 * Problem: Can not target requests because IBusTopicReplyer creates it own
 * hidden subscriber that has no message selector.
 */
class ReceiverReplyer extends IBusTopicReplyer implements GMS_MessageReplyer {

  private GMS_MessageListener m_listener;
  private GMS_SubscriberImpl  m_parent;
  private TopicSubscriber     m_subscriber; // For receiving requests.
  private TopicPublisher      m_publisher; // For replies.
  private MessageListener     m_listenerAdapter; // Adapts m_listener to javax.jms.MessageListener

  ReceiverReplyer(GMS_SubscriberImpl parent,GMS_MessageListener l) throws JMSException {
    super(parent.m_subscriber.getTopic(),
          IBusJMSContext.createTopicSession(false,TopicSession.AUTO_ACKNOWLEDGE));
    m_parent = parent;
    m_listener = l;
    m_publisher = getSession().createPublisher(m_parent.m_subscriber.getTopic());
    m_listenerAdapter = new IBusMessageListenerAdapter(l,parent);
  }

  public void reply(GMS_Message original,GMS_Message reply) throws GXO_Exception {
    try {
      Message originalM = GMS_PublisherImpl.convertMessage(original);
      Message replyM    = GMS_PublisherImpl.convertMessage(reply);
      Destination replyTo = originalM.getJMSReplyTo();

      reply(m_publisher,originalM,replyM);

      LogManager.log(LogManager.ALL,"Receiver","Sent reply " + replyM);
    } catch (JMSException e) {
      throw new GXO_JMSException(e);
    }
  }

  public void onRequest(Message m) {
   try {
      if (m.getJMSReplyTo() == null) {
        LogManager.log(LogManager.ALL,"Receiver","Got Message: [" +
            m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+m);
        m_listenerAdapter.onMessage(m);
      } else {
        LogManager.log(LogManager.ALL,"Receiver","Got Request: [" +
              m.getStringProperty(GXO_JMSConstants.GXO_DESTINATION_PROPERTY)+
                          "] [correlation: "+m.getJMSCorrelationID()+"]"+m);
        m_listener.onRequest(GMS_SubscriberImpl.convertMessage(m),this);
      }
    } catch (Exception e) {
      LogManager.err("Subscriber","Received unsupported message type",e);
    }
  }
}
