package com.anabas.commsharedlet;

import com.anabas.sharedlet.*;
import com.anabas.naming.*;
import com.anabas.util.misc.LogManager;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Title:        Communication sharedlet
 * Description:  This project contains the Anabas, Inc. collaboration framework libraries.
 * Copyright:    Copyright (c) 2000
 * Company:      Anabas, Inc.
 * @author
 * @version 1.0
 *
 * This is an implementation of Sharedlet class that is used by
 * DefaultSharedletManager to store sharedlet session logic & view information.
 *
 * This class is instantiated only by DefaultSharedletManager and will "lazy"
 * instantiate views & logic as needed.
 *
 * It is the access point for containers to get sharedlet view & session
 * logic instances.
 */

public class CommunicationSharedlet implements Sharedlet {

  private SharedletManager m_sharedletManager;
  private SharedletViewManager m_viewManager;
  private SharedletLogicManager m_logicManager;
  private String m_MIMEType;

  /**
   * Should only be invoked by DefaultSharedletManager.
   */
  public CommunicationSharedlet() {
    m_MIMEType = "application/x-sharedlet-communication";
  }

  /**
   * Find the view instance inside the view manager and return it.  This function
   * is design so that it "lazy" instantiates.
   *
   * @param id The id returned by SharedletViewInfo.getID().
   * @return The sharedlet view associated with the particular id.  null if the
   *         view does not exist.
   */
  public SharedletView getView(String id) {
    if (m_sharedletManager == null) init();

    SharedletInfo sharedletInfo = m_sharedletManager.getInfo(this);
    Vector v = sharedletInfo.getViewInfos();

    Enumeration enum = v.elements();
    while(enum.hasMoreElements()) {
      SharedletViewInfo vinfo = (SharedletViewInfo)enum.nextElement();
      if (vinfo.getID().equals(id)) {
        return m_viewManager.getView(vinfo);
      }
    }
    return null;
  }

  /**
   * Proxy function for sharedlet manager's getSessionLogic().
   *
   * @param id The id returned by SharedletSessionLogic.getID().
   */
  public SharedletSessionLogic getSessionLogic(String id) {
    if (m_sharedletManager == null) init();

    SharedletInfo sharedletInfo = m_sharedletManager.getInfo(this);
    Vector v = sharedletInfo.getLogicInfos();

    Enumeration enum = v.elements();
    while(enum.hasMoreElements()) {
      SharedletSessionLogicInfo linfo = (SharedletSessionLogicInfo)enum.nextElement();
      if (linfo.getID().equals(id)) {
        return m_logicManager.getLogic(linfo);
      }
    }
    return null;
  }

  /**
   * The default sharedlet doesn't support any document types....so no saving.
   */
  public void loadDocument(java.net.URL url) {
    LogManager.err(m_MIMEType,"Trying to load " + url +
                   " but doesn't support the document type");
  }

  /**
   * Grabs the handle to the necessary managers.
   */
  private void init() {
    Context ctx = ContextManager.getInitialContext();
    try {
      m_sharedletManager = (SharedletManager)ctx.lookup("services/SharedletManager");
      m_viewManager = m_sharedletManager.getViewManager();
      m_logicManager = m_sharedletManager.getLogicManager();
    } catch (NamingException e) {
      LogManager.err("SharedletImpl","Unable to get handle to sharedlet manager",e);
    }

  }
}